//////////////////////////////////////////////////////////////////////////////
//
// Copyright (c) 2005
//   Nicholas Kidd  <kidd@cs.wisc.edu>
//   Thomas Reps    <reps@cs.wisc.edu>
//   David Melski   <melski@grammatech.com>
//   Akash Lal      <akash@cs.wisc.edu>
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   1. Redistributions of source code must retain the above copyright
//      notice, this list of conditions and the following disclaimer.
//
//   2. Redistributions in binary form must reproduce the above copyright
//      notice, this list of conditions and the following disclaimer in
//      the documentation and/or other materials provided with the
//      distribution.
//
//   3. Neither the name of the University of Wisconsin, Madison,
//      nor GrammaTech, Inc., nor the names of the copyright holders
//      and contributors may be used to endorse or promote
//      products derived from this software without specific prior
//      written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
// A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//////////////////////////////////////////////////////////////////////////////
/*
 * $Id: dictionary.h,v 1.20 2005/06/22 22:05:11 kidd Exp $
 */
#ifndef wpds_DICTIONARY_H_
#define wpds_DICTIONARY_H_
#include <vector>
#include "common.h"
#include "HashMap.h"
#include "key_source.h"

namespace wpds {

	class Dictionary {

		public:		// Constructor/DE

            /*!
             * Create dictionary reserving buckets space for the map.
             */
			Dictionary( wpds_size_t buckets=6151 );

			// Destructor must delete all the "dup'd"
			// strings stored in values
			~Dictionary();

		public:		// methods

			// only use add_new_item if you are
            // positive it is a new item.
            // When there is an existing item for s, add_item reuses it and
            // "delete" s.
			wpds_key_t add_item( key_source * s);
            wpds_key_t add_new_item( key_source * s );
            void traverse(void (*f)(key_source *));
            key_source *retrieve_item( wpds_key_t );
			// Does not touch memory of s
			bool exists(key_source * s);

            void clear();

			wpds_size_t idx() const { return values.size(); }

            wpds_size_t num_keys() const { return values.size(); }

		private:	// typedef
            typedef HashMap< const key_source *,wpds_key_t > map_type;
			typedef std::vector< key_source * > vector_type;

		private:	// vars
			// Maps a hash to a set of keys
			map_type keys;
			vector_type values;
	};

}
#endif	// wpds_DICTIONARY_H_

/* Yo, Emacs!
;;; Local Variables: ***
;;; tab-width: 4 ***
;;; End: ***
*/
